# _moflextab.py. This file automatically created by PLY (version 3.11). Don't edit!
_tabversion   = '3.10'
_lextokens    = set(('ANY', 'AS', 'ASSOCIATION', 'CLASS', 'DISABLEOVERRIDE', 'DT_BOOL', 'DT_CHAR16', 'DT_DATETIME', 'DT_REAL32', 'DT_REAL64', 'DT_SINT16', 'DT_SINT32', 'DT_SINT64', 'DT_SINT8', 'DT_STR', 'DT_UINT16', 'DT_UINT32', 'DT_UINT64', 'DT_UINT8', 'ENABLEOVERRIDE', 'FALSE', 'FLAVOR', 'IDENTIFIER', 'INDICATION', 'INSTANCE', 'METHOD', 'NULL', 'OF', 'PARAMETER', 'PRAGMA', 'PROPERTY', 'QUALIFIER', 'REF', 'REFERENCE', 'RESTRICTED', 'SCHEMA', 'SCOPE', 'TOINSTANCE', 'TOSUBCLASS', 'TRANSLATABLE', 'TRUE', 'binaryValue', 'charValue', 'decimalValue', 'floatValue', 'hexValue', 'octalValue', 'stringValue'))
_lexreflags   = 64
_lexliterals  = '#(){};[],$:='
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_COMMENT>//.*)|(?P<t_MCOMMENT>/\\*(.|\\n)*?\\*/)|(?P<t_floatValue>[+-]?[0-9]*\\.[0-9]+([eE][+-]?[0-9]+)?)|(?P<t_hexValue>[+-]?0[xX][0-9a-fA-F]+)|(?P<t_binaryValue>[+-]?[0-9]+[bB])|(?P<t_octalValue>[+-]?0[0-9]+)|(?P<t_decimalValue>[+-]?([1-9][0-9]*|0))|(?P<t_charValue>\'([^\'\\\\\\n\\r]|([\\\\](([bfnrt\'"\\\\])|([xX][0-9a-fA-F]{1,4}))))\')|(?P<t_stringValue>"([^"\\\\\\n\\r]|([\\\\](([bfnrt\'"\\\\])|([xX][0-9a-fA-F]{1,4}))))*")|(?P<t_IDENTIFIER>([a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))([0-9a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))*)|(?P<t_newline>\\n+)', [None, ('t_COMMENT', 'COMMENT'), ('t_MCOMMENT', 'MCOMMENT'), None, ('t_floatValue', 'floatValue'), None, ('t_hexValue', 'hexValue'), ('t_binaryValue', 'binaryValue'), ('t_octalValue', 'octalValue'), ('t_decimalValue', 'decimalValue'), None, ('t_charValue', 'charValue'), None, None, None, None, None, ('t_stringValue', 'stringValue'), None, None, None, None, None, ('t_IDENTIFIER', 'IDENTIFIER'), None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, ('t_newline', 'newline')])]}
_lexstateignore = {'INITIAL': ' \r\t'}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}
